import base64
import time
import traceback
from threading import Event, Thread
import ssl
import json
import logging
from pyee import ExecutorEventEmitter
from websocket import WebSocketApp, WebSocketConnectionClosedException,  \
    WebSocketException
from hivemind_bus_client.message import HiveMessage, HiveMessageType
from hivemind_bus_client.util import serialize_message, \
    encrypt_as_json, decrypt_from_json
from mycroft_bus_client import Message as MycroftMessage


LOG = logging.getLogger("HiveMind-websocket-client")


class HiveMessageWaiter:
    """Wait for a single message.

    Encapsulate the wait for a message logic separating the setup from
    the actual waiting act so the waiting can be setuo, actions can be
    performed and _then_ the message can be waited for.

    Argunments:
        bus: Bus to check for messages on
        message_type: message type to wait for
    """

    def __init__(self, bus, message_type):
        self.bus = bus
        self.msg_type = message_type
        self.received_msg = None
        # Setup response handler
        self.response_event = Event()
        self.bus.once(message_type, self._handler)

    def _handler(self, message):
        """Receive response data."""
        self.received_msg = message
        self.response_event.set()

    def wait(self, timeout=3.0):
        """Wait for message.

        Arguments:
            timeout (int or float): seconds to wait for message

        Returns:
            HiveMessage or None
        """
        self.response_event.wait(timeout)
        if not self.response_event.is_set():
            # Clean up the event handler
            try:
                self.bus.remove(self.msg_type, self._handler)
            except (ValueError, KeyError):
                # ValueError occurs on pyee 5.0.1 removing handlers
                # registered with once.
                # KeyError may theoretically occur if the event occurs as
                # the handler is removed
                pass
        return self.received_msg


class HivePayloadWaiter(HiveMessageWaiter):
    def __init__(self, payload_type=HiveMessageType.THIRDPRTY, *args, **kwargs):
        super(HivePayloadWaiter, self).__init__(*args, **kwargs)
        self.payload_type = payload_type

    def _handler(self, message):
        """Receive response data."""
        if message.payload.msg_type == self.payload_type:
            self.received_msg = message
            self.response_event.set()
        else:
            self.bus.once(self.msg_type, self._handler)


class HiveMessageBusClient:
    def __init__(self, key, crypto_key=None, host='127.0.0.1', port=5678,
                 useragent="HiveMessageBusClientV0.0.1", ssl=True,
                 self_signed=False, debug=False):
        self.ssl = ssl
        self.key = key
        self.useragent = useragent
        self.crypto_key = crypto_key
        self.host = host
        self.port = port
        self.emitter = ExecutorEventEmitter()
        self.client = self.create_client()
        self.retry = 5
        self.connected_event = Event()
        self.started_running = False
        self.allow_self_signed = self_signed
        self._mycroft_events = {}  # msg_type: [handler]
        self.debug = debug

    @staticmethod
    def build_url(key, host='127.0.0.1', port=5678,
                  useragent="HiveMessageBusClientV0.0.1", ssl=True):
        scheme = 'wss' if ssl else 'ws'
        key = base64.b64encode(f"{useragent}:{key}".encode("utf-8"))\
            .decode("utf-8")
        return f'{scheme}://{host}:{port}?authorization={key}'

    def create_client(self):
        url = self.build_url(ssl=self.ssl,
                             host=self.host,
                             port=self.port,
                             key=self.key,
                             useragent=self.useragent)
        return WebSocketApp(url, on_open=self.on_open, on_close=self.on_close,
                            on_error=self.on_error,
                            on_message=self.on_message)

    def run_in_thread(self):
        """Launches the run_forever in a separate daemon thread."""
        t = Thread(target=self.run_forever)
        t.daemon = True
        t.start()
        return t

    def run_forever(self):
        self.started_running = True
        if self.allow_self_signed:
            self.client.run_forever(sslopt={
                "cert_reqs": ssl.CERT_NONE,
                "check_hostname": False,
                "ssl_version": ssl.PROTOCOL_TLSv1})
        else:
            self.client.run_forever()

    def close(self):
        self.client.close()
        self.connected_event.clear()

    # event handlers
    def on_open(self):
        LOG.info("Connected")
        self.connected_event.set()
        self.emitter.emit("open")
        # Restore reconnect timer to 5 seconds on sucessful connect
        self.retry = 5

    def on_close(self):
        self.emitter.emit("close")

    def on_error(self, error):
        """ On error start trying to reconnect to the websocket. """
        if isinstance(error, WebSocketConnectionClosedException):
            LOG.warning('Could not send message because connection has closed')
        else:
            LOG.exception('=== ' + repr(error) + ' ===')

        try:
            self.emitter.emit('error', error)
            if self.client.keep_running:
                self.client.close()
        except Exception as e:
            LOG.error('Exception closing websocket: ' + repr(e))

        LOG.warning(
            "HiveMessage Bus Client will reconnect in %d seconds." % self.retry
        )
        time.sleep(self.retry)
        self.retry = min(self.retry * 2, 60)
        try:
            self.emitter.emit('reconnecting')
            self.client = self.create_client()
            self.run_forever()
        except WebSocketException:
            pass

    def on_message(self, message):
        if self.crypto_key:
            if "ciphertext" in message:
                message = decrypt_from_json(self.crypto_key,  message)
            else:
                LOG.warning("Message was unencrypted")
        if isinstance(message, str):
            message = json.loads(message)
        if self.debug:
            print("received: ", message)
        self.emitter.emit('message', message)
        parsed_message = HiveMessage(**message)
        self.emitter.emit(parsed_message.msg_type, parsed_message)
        if parsed_message.msg_type == HiveMessageType.BUS:
            self._fire_mycroft_handlers(parsed_message)

    def emit(self, message):
        if not self.connected_event.wait(10):
            if not self.started_running:
                raise ValueError('You must execute run_forever() '
                                 'before emitting messages')
            self.connected_event.wait()

        if isinstance(message, MycroftMessage):
            message = HiveMessage(msg_type=HiveMessageType.BUS,
                                  payload=message)
        try:
            # auto inject context for proper routing, this is confusing for
            # end users if they need to do it manually, error prone and easy
            # to forget
            if message.msg_type == HiveMessageType.BUS:
                ctxt = message.payload.context
                if "source" not in ctxt :
                    ctxt["source"] = self.useragent
                if "platform" not in message.payload.context:
                    ctxt["platform"] = self.useragent
                if "destination" not in message.payload.context:
                    ctxt["destination"] = "JarbasHiveMind"
                message._payload["context"] = ctxt
            payload = serialize_message(message)
            if self.crypto_key:
                payload = encrypt_as_json(self.crypto_key, payload)

            self.client.send(payload)
        except WebSocketConnectionClosedException:
            LOG.warning('Could not send {} message because connection '
                        'has been closed'.format(message.msg_type))

    # mycroft events api
    def _fire_mycroft_handlers(self, message):
        handlers = []
        if isinstance(message, MycroftMessage):
            handlers = self._mycroft_events.get(message.msg_type) or []
        elif message.msg_type == HiveMessageType.BUS:
            handlers = self._mycroft_events.get(message.payload.msg_type) or []
        for func in handlers:
            try:
                func(message.payload)
            except Exception as e:
                if self.debug:
                    print(e)
                continue

    def emit_mycroft(self, message):
        message = HiveMessage(msg_type=HiveMessageType.BUS, payload=message)
        self.emit(message)

    def on_mycroft(self, mycroft_msg_type, func):
        self._mycroft_events[mycroft_msg_type] = self._mycroft_events.get(mycroft_msg_type) or []
        self._mycroft_events[mycroft_msg_type].append(func)

    # event api
    def on(self, event_name, func):
        if event_name not in list(HiveMessageType):
            # assume it's a mycroft message
            # this could be done better,
            # but makes this lib almost a drop in replacement
            # for the mycroft bus client
            self.on_mycroft(event_name, func)
        else:
            # hivemind message
            self.emitter.on(event_name, func)

    def once(self, event_name, func):
        self.emitter.once(event_name, func)

    def remove(self, event_name, func):
        try:
            if not event_name in self.emitter._events:
                LOG.debug("Not able to find '" + str(event_name) + "'")
            self.emitter.remove_listener(event_name, func)
        except ValueError:
            LOG.warning('Failed to remove event {}: {}'.format(event_name,
                                                               str(func)))
            for line in traceback.format_stack():
                LOG.warning(line.strip())

            if not event_name in self.emitter._events:
                LOG.debug("Not able to find '" + str(event_name) + "'")
            LOG.warning("Existing events: " + str(self.emitter._events))
            for evt in self.emitter._events:
                LOG.warning("   " + str(evt))
                LOG.warning("       " + str(self.emitter._events[evt]))
            if event_name in self.emitter._events:
                LOG.debug("Removing found '" + str(event_name) + "'")
            else:
                LOG.debug("Not able to find '" + str(event_name) + "'")
            LOG.warning('----- End dump -----')

    def remove_all_listeners(self, event_name):
        """Remove all listeners connected to event_name.

            Arguments:
                event_name: event from which to remove listeners
        """
        if event_name is None:
            raise ValueError
        self.emitter.remove_all_listeners(event_name)

    # utility
    def wait_for_message(self, message_type, timeout=3.0):
        """Wait for a message of a specific type.

        Arguments:
            message_type (HiveMessageType): the message type of the expected message
            timeout: seconds to wait before timeout, defaults to 3

        Returns:
            The received message or None if the response timed out
        """

        return HiveMessageWaiter(self, message_type).wait(timeout)

    def wait_for_payload(self, payload_type,
                         message_type=HiveMessageType.THIRDPRTY,
                         timeout=3.0):
        """Wait for a message of a specific type + payload of a specific type.

        Arguments:
            payload_type (str): the message type of the expected payload
            message_type (HiveMessageType): the message type of the expected message
            timeout: seconds to wait before timeout, defaults to 3

        Returns:
            The received message or None if the response timed out
        """

        return HivePayloadWaiter(bus=self, payload_type=payload_type,
                                 message_type=message_type).wait(timeout)

    def wait_for_mycroft(self, mycroft_msg_type, timeout=3.0):
        return self.wait_for_payload(mycroft_msg_type, timeout=timeout,
                                     message_type=HiveMessageType.BUS)

    def wait_for_response(self, message, reply_type=None, timeout=3.0):
        """Send a message and wait for a response.

        Arguments:
            message (HiveMessage): message to send, mycroft Message objects also accepted
            reply_type (HiveMessageType): the message type of the expected reply.
                                          Defaults to "<message.msg_type>".
            timeout: seconds to wait before timeout, defaults to 3

        Returns:
            The received message or None if the response timed out
        """
        if isinstance(message, MycroftMessage):
            message = HiveMessage(msg_type=HiveMessageType.BUS, payload=message)
        message_type = reply_type or message.msg_type
        waiter = HiveMessageWaiter(self, message_type)  # Setup response handler
        # Send message and wait for it's response
        self.emit(message)
        return waiter.wait(timeout)

    def wait_for_payload_response(self, message, payload_type,
                                  reply_type=None, timeout=3.0):
        """Send a message and wait for a response.

        Arguments:
            message (HiveMessage): message to send, mycroft Message objects also accepted
            payload_type (str): the message type of the expected payload
            reply_type (HiveMessageType): the message type of the expected reply.
                                          Defaults to "<message.msg_type>".
            timeout: seconds to wait before timeout, defaults to 3

        Returns:
            The received message or None if the response timed out
        """
        if isinstance(message, MycroftMessage):
            message = HiveMessage(msg_type=HiveMessageType.BUS, payload=message)
        message_type = reply_type or message.msg_type
        waiter = HivePayloadWaiter(bus=self, payload_type=payload_type,
                                   message_type=message_type)  # Setup
        # response handler
        # Send message and wait for it's response
        self.emit(message)
        return waiter.wait(timeout)


from enum import Enum
import json
from ovos_utils.json_helper import merge_dict
from mycroft_bus_client import Message


class HiveMessageType(str, Enum):
    HANDSHAKE = "shake"  # negotiate initial connection
    BUS = "bus"  # request meant for internal mycroft-bus in master
    SHARED_BUS = "shared_bus"  # passive sharing of message
                               # from mycroft-bus in slave
    BROADCAST = "broadcast"  # forward message to all slaves
    PROPAGATE = "propagate"  # forward message to all slaves and masters
    ESCALATE = "escalate"  # forward message up the authority chain to all
                           # masters
    HELLO = "hello"  # like escalate, used to announce the device
    QUERY = "query"  # like escalate, but stops once one of the nodes can
                     # send a response
    CASCADE = "cascade"  # like propagate, but expects a response back from
                         # all nodes in the hive (responses optional)
    PING = "ping"  # like cascade, but used to map the network
    RENDEZVOUS = "rendezvous"  # reserved for rendezvous-nodes
    THIRDPRTY = "3rdparty"  # user land message, do whatever you want


class HiveMessage:
    def __init__(self, msg_type, payload=None, node=None, source_peer=None,
                 route=None, target_peers=None):
        #  except for the hivemind node classes receiving the message and
        #  creating the object nothing should be able to change these values
        #  node classes might change them a runtime by the private attribute
        #  but end-users should consider them read_only
        if msg_type not in [m.value for m in HiveMessageType]:
            raise ValueError("Unknown HiveMessage.msg_type")

        self._msg_type = msg_type
        # the payload is more or less a free for all
        # the msg_type determines what happens to the message, but the
        # payload can simply be ignored by the receiving module

        # some msg_types might return HiveMessage, others (mycroft) Message
        # we should support the dict/json format, json is used at the
        # transport layer before converting into any of these formats
        if isinstance(payload, Message):
            payload = {"type": payload.msg_type,
                       "data": payload.data,
                       "context": payload.context}
        elif isinstance(payload, str):
            payload = json.loads(payload)
        self._payload = payload or {}

        self._node = node  # node semi-unique identifier
        self._source_peer = source_peer  # peer_id
        self._route = route or []  # where did this message come from
        self._targets = target_peers or []  # where will it be sent

    @property
    def msg_type(self):
        return self._msg_type

    @property
    def node_id(self):
        return self._node

    @property
    def source_peer(self):
        return self._source_peer

    @property
    def target_peers(self):
        if self.source_peer:
            return self._targets or [self._source_peer]
        return self._targets

    @property
    def route(self):
        return [r for r in self._route if r.get("targets") and r.get("source")]

    @property
    def payload(self):
        if self.msg_type in [HiveMessageType.BUS, HiveMessageType.SHARED_BUS]:
            return Message(self._payload["type"],
                           data=self._payload.get("data"),
                           context=self._payload.get("context"))
        if self.msg_type in [HiveMessageType.BROADCAST,
                             HiveMessageType.PROPAGATE,
                             HiveMessageType.CASCADE,
                             HiveMessageType.ESCALATE]:
            return HiveMessage(**self._payload)
        return self._payload

    @property
    def as_dict(self):
        pload = self._payload
        if isinstance(pload, HiveMessage):
            pload = pload.as_json
        elif isinstance(pload, Message):
            pload = pload.serialize()
        if isinstance(pload, str):
            pload = json.loads(pload)
        return {"msg_type": self.msg_type,
                "payload": pload,
                "route": self.route,
                "node": self.node_id,
                "source_peer": self.source_peer}

    @property
    def as_json(self):
        return json.dumps(self.as_dict)

    def serialize(self):
        return self.as_json

    def __getitem__(self, item):
        return self._payload.get(item)

    def __setitem__(self, key, value):
        self._payload[key] = value

    def __str__(self):
        return self.as_json

    def update_hop_data(self, data=None, **kwargs):
        if not self._route or self._route[-1]["source"] != self.source_peer:
            self._route += [{"source": self.source_peer,
                             "targets": self.target_peers}]
        if self._route and data:
            self._route[-1] = merge_dict(self._route[-1], data, **kwargs)

    def replace_route(self, route):
        self._route = route

    def update_source_peer(self, peer):
        self._source_peer = peer
        return self

    def add_target_peer(self, peer):
        self._targets.append(peer)

    def remove_target_peer(self, peer):
        if peer in self._targets:
            self._targets.pop(peer)



import json
from ovos_utils.security import encrypt, decrypt
from binascii import hexlify, unhexlify


def serialize_message(message):
    # convert a Message object into raw data that can be sent over
    # websocket
    if hasattr(message, 'serialize'):
        return message.serialize()
    elif isinstance(message, dict):
        message = {
            k: v if not hasattr(v, 'serialize') else serialize_message(v)
            for k, v in message.items()}
        return json.dumps(message)
    else:
        return json.dumps(message.__dict__)


def encrypt_as_json(key, data, nonce=None):
    if isinstance(data, dict):
        data = json.dumps(data)
    if len(key) > 16:
        key = key[0:16]
    try:
        ciphertext, tag, nonce = encrypt(key, data, nonce=nonce)
    except:
        raise RuntimeError("EncryptionKeyError")
    return json.dumps({"ciphertext": hexlify(ciphertext).decode('utf-8'),
                       "tag": hexlify(tag).decode('utf-8'),
                       "nonce": hexlify(nonce).decode('utf-8')})


def decrypt_from_json(key, data):
    if isinstance(data, str):
        data = json.loads(data)
    if len(key) > 16:
        key = key[0:16]
    ciphertext = unhexlify(data["ciphertext"])
    if data.get("tag") is None:  # web crypto
        ciphertext, tag = ciphertext[:-16], ciphertext[-16:]
    else:
        tag = unhexlify(data["tag"])
    nonce = unhexlify(data["nonce"])
    try:
        return decrypt(key, ciphertext, tag, nonce)
    except ValueError:
        raise RuntimeError("DecryptionKeyError")



from hivemind_bus_client.message import HiveMessageType


class HiveMessageListener:
    def __init__(self, bus, message_type):
        self.bus = bus
        self.message_type = message_type
        self._handlers = []

    def _handler(self, message):
        """Receive response data."""
        for handler in self._handlers:
            handler(message)
        self.bus.once(self.message_type, self._handler)

    def listen(self):
        self.bus.once(self.message_type, self._handler)
        return self

    def add_handler(self, handler):
        self._handlers.append(handler)

    def clear_handlers(self):
        self._handlers = []

    def shutdown(self):
        self.bus.remove(self.message_type, self._handler)


class HivePayloadListener(HiveMessageListener):
    def __init__(self, payload_type=HiveMessageType.THIRDPRTY, *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.payload_type = payload_type

    def _handler(self, message):
        """Receive response data."""
        if message.payload.msg_type == self.payload_type:
            for handler in self._handlers:
                handler(message.payload)
        self.bus.once(self.message_type, self._handler)


def on_hive_message(message_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        bus.on(message_type, func)
        return func

    return wrapped_handler


def on_mycroft_message(payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=HiveMessageType.BUS)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


def on_shared_bus(payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=HiveMessageType.SHARED_BUS)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


def on_broadcast(payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=HiveMessageType.BROADCAST)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


def on_ping(payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=HiveMessageType.PING)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


def on_propagate(payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=HiveMessageType.PROPAGATE)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


def on_escalate(payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=HiveMessageType.ESCALATE)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


def on_handshake(payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=HiveMessageType.HANDSHAKE)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


def on_hello(payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=HiveMessageType.HELLO)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


def on_cascade(payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=HiveMessageType.CASCADE)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


def on_rendezvous(payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=HiveMessageType.RENDEZVOUS)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


def on_registry_opcode(payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=HiveMessageType.THIRDPRTY)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


# low level
def on_third_party(bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HiveMessageListener(bus=bus,
                                     message_type=HiveMessageType.THIRDPRTY)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler


def on_payload(message_type, payload_type, bus):
    # Begin wrapper
    def wrapped_handler(func):
        waiter = HivePayloadListener(bus=bus, payload_type=payload_type,
                                     message_type=message_type)
        waiter.add_handler(func)
        waiter.listen()
        func.shutdown = waiter.shutdown
        return func

    return wrapped_handler



import os
from setuptools import setup


with open("README.md", "r") as fh:
    long_desc = fh.read()


def required(requirements_file):
    """ Read requirements file and remove comments and empty lines. """
    base_dir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(base_dir, requirements_file), 'r') as f:
        requirements = f.read().splitlines()
        return [pkg for pkg in requirements
                if pkg.strip() and not pkg.startswith("#")]


setup(
    name='hivemind_bus_client',
    version='0.0.3a1',
    packages=['hivemind_bus_client'],
    package_data={
      '*': ['*.txt', '*.md']
    },
    include_package_data=True,
    install_requires=required('requirements.txt'),
    url='https://github.com/JarbasHiveMind/hivemind_websocket_client',
    license='Apache-2.0',
    author='JarbasAi',
    author_email='jarbasai@mailfence.com',
    description='Hivemind Websocket Client',
    long_description=long_desc,
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)
